# coding: utf-8
from __future__ import absolute_import, unicode_literals
from builtins import object
from gqlspection import log
import gqlspection
from .utils import safe_get_list


class GQLWrapFactory(object):
    schema = None
    json   = None

    def __init__(self, schema, json):
        self.schema = schema
        self.json = json

    def fields(self):
        return GQLFields(self.schema, self.json)

    def interfaces(self):
        return GQLInterfaces(self.schema, self.json)

    def enums(self):
        return GQLEnums(self.schema, self.json)

    def args(self):
        return GQLArgs(self.schema, self.json)


class GQLWrapper(gqlspection.GQLList):
    # This should be overwritten through inheritance
    @staticmethod
    def _extract_elements(schema, json):
        return ()

    def __init__(self, schema, json):
        elements = (
            element for element in self._extract_elements(schema, json)
            if not element.name.startswith('__')
        )

        super(GQLWrapper, self).__init__(elements)


class GQLTypes(GQLWrapper):
    def _extract_elements(self, schema, json):
        elements = []
        scalars = set()
        for t in safe_get_list(json, 'types'):
            el = gqlspection.GQLType.from_json(t, schema)
            log.info("Adding new type definition: %s", el.name)
            elements.append(el)

            if el.kind.kind == 'SCALAR':
                scalars.add(el.kind.name)

        # populate standard types if not present in supplied schema
        for scalar in gqlspection.GQLTypeKind.builtin_scalars:
            log.info("Adding missing default scalar: %s" % scalar)
            if scalar not in scalars:
                elements.append(gqlspection.GQLType(
                    name = scalar,
                    kind = gqlspection.GQLTypeKind(
                        name=scalar,
                        kind='SCALAR'
                    ),
                    description="Built-in scalar type.",
                    schema=schema
                ))

        return elements


class GQLFields(GQLWrapper):
    @staticmethod
    def _extract_elements(schema, json):
        return (gqlspection.GQLField.from_json(field, schema) for field in safe_get_list(json, 'fields'))


class GQLInterfaces(GQLWrapper):
    """Iterator-like wrapper around 'interfaces' type attribute which lists interfaces that object implements.

    An example of a valid field:

          "interfaces": [
            {
              "kind": "INTERFACE",
              "name": "Error",
              "ofType": null
            }
          ],
    """
    @staticmethod
    def _extract_elements(schema, json):
        return (gqlspection.GQLTypeProxy(interface['name'], schema) for interface in safe_get_list(json, 'interfaces'))


class GQLEnums(GQLWrapper):
    @staticmethod
    def _extract_elements(schema, json):
        return (gqlspection.GQLEnum.from_json(enum) for enum in safe_get_list(json, 'enumValues'))


class GQLArgs(GQLWrapper):
    @staticmethod
    def _extract_elements(schema, json):
        return (gqlspection.GQLArg.from_json(arg, schema) for arg in safe_get_list(json, 'inputFields'))



# coding: utf-8
from __future__ import unicode_literals
from gqlspection.Logger import log

from gqlspection.GQLArg import GQLArg
from gqlspection.GQLEnum import GQLEnum
from gqlspection.GQLField import GQLField
from gqlspection.GQLList import GQLList
from gqlspection.GQLQuery import GQLQuery
from gqlspection.GQLSchema import GQLSchema
from gqlspection.GQLSubQuery import GQLSubQuery
from gqlspection.GQLType import GQLType
from gqlspection.GQLTypeKind import GQLTypeKind
from gqlspection.GQLTypeProxy import GQLTypeProxy
from gqlspection.GQLWrappers import GQLWrapFactory, GQLArgs, GQLEnums, GQLInterfaces, GQLFields, GQLTypes

__all__ = [
    "log",
    "utils",
    "GQLArg",
    "GQLEnum",
    "GQLField",
    "GQLList",
    "GQLQuery",
    "GQLSchema",
    "GQLSubQuery",
    "GQLType",
    "GQLTypeKind",
    "GQLTypeProxy",
    "GQLWrapFactory",
    "GQLArgs",
    "GQLEnums",
    "GQLInterfaces",
    "GQLFields",
    "GQLTypes"
]

# CLI tool entrypoint
from platform import python_implementation
if python_implementation() != 'Jython':
    from gqlspection.cli import cli  # noqa
    __all__.append("cli")



# coding: utf-8
from __future__ import unicode_literals
from builtins import object


class GQLEnum(object):
    name = ''
    description = ''
    is_deprecated = False
    deprecation_reason = ''

    def __init__(self, name, description = '', is_deprecated=False, deprecation_reason=''):
        self.name = name
        self.description = description
        self.is_deprecated = is_deprecated
        self.deprecation_reason = deprecation_reason

    @staticmethod
    def from_json(json):
        return GQLEnum(
            name=json['name'],
            description=json.get('description', ''),
            is_deprecated=json.get('is_deprecated', False),
            deprecation_reason=json.get('deprecation_reason', '')
        )

    def __repr__(self):
        return '{name} ({description})'.format(
            name = self.name,
            description = self.description
        )



# coding: utf-8
from __future__ import unicode_literals
from builtins import str, object
import gqlspection


class GQLType(object):
    name = ''
    kind = None
    schema = None
    description = ''
    fields = None
    interfaces = None
    enums = None
    args = None
    url = ''

    def __init__(self, name, kind, schema, description='', fields=None, interfaces=None, enums=None, args=None, url=''):
        self.name = name
        self.kind = kind
        self.schema = schema
        # Optional strings
        self.description = description
        self.url         = url
        # Optional GQLWrappers
        self.args       = args       or gqlspection.GQLArgs(schema, {})
        self.fields     = fields     or gqlspection.GQLFields(schema, {})
        self.interfaces = interfaces or gqlspection.GQLInterfaces(schema, {})
        self.enums      = enums      or gqlspection.GQLEnums(schema, {})

    @staticmethod
    def from_json(json, schema):
        wrap = gqlspection.GQLWrapFactory(schema, json)

        return GQLType(
            name=json['name'],
            kind=gqlspection.GQLTypeKind.from_json(json),
            schema=schema,
            description=json.get('description', ''),

            fields     = wrap.fields(),
            interfaces = wrap.interfaces(),
            enums      = wrap.enums(),
            args       = wrap.args(),

            url=json.get('specifiedByURL', '')
        )

    def __repr__(self):
        return '"{name}" ({kind}) - {description} - [fields: {fields}] [interfaces: {interfaces}] [enums: {enums}] [ ' \
               'args: {args}]'.format(
            name        = self.name,
            kind        = self.kind.kind,
            description = self.description,
            fields      = ', '.join([str(x) for x in self.fields]),
            interfaces  = ', '.join([str(x) for x in self.interfaces]),
            enums       = ', '.join([str(x) for x in self.enums]),
            args        = ', '.join([str(x) for x in self.args])
        )



# coding: utf-8
from __future__ import unicode_literals
from builtins import object
import logging


class Logger(object):
    logger = None

    def __init__(self):
        logging.basicConfig()
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.WARNING)

    def debug(self, msg, *args, **kwargs):
        self.logger.debug(msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self.logger.info(msg, *args, **kwargs)

    def warn(self, msg, *args, **kwargs):
        self.logger.warning(msg, *args, **kwargs)

    def err(self, msg, *args, **kwargs):
        self.logger.error(msg, *args, **kwargs)

    @property
    def is_debug(self):
        return self.logger.level <= logging.DEBUG


log = Logger()



# coding: utf-8
from __future__ import print_function, unicode_literals
import click
import json
import sys
try:
    from pathlib import Path
except ImportError:
    from pathlib2 import Path
from gqlspection import log, GQLSchema

click.disable_unicode_literals_warning = True

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option(
    '-f', '--file', 'file_', help="File with the GraphQL schema (introspection JSON)."
)
@click.option(
    '-u', '--url', help="URL of the GraphQL endpoint with enabled introspection."
)
@click.option(
    '-l', '--list', 'stuff_to_print', help="Parse GraphQL schema and list queries, mutations or both of them (valid "
                                           "values are: 'queries', 'mutations' or 'all')."
)
@click.option(
    '-q', '--query', help="Only print named queries (argument is a comma-separated list of query names)."
)
@click.option(
    '-m', '--mutation', help="Only print named mutations (argument is a comma-separated list of mutation names)."
)
@click.option(
    '-Q', '--all-queries', is_flag=True, help="Only print queries (by default both queries and mutations are printed)."
)
@click.option(
    '-M', '--all-mutations', is_flag=True, help="Only print mutations (by default both queries and mutations are "
                                                "printed)."
)
@click.option(
    '-v', '--verbose', is_flag=True, help="Enable verbose logging."
)
def cli(file_=None, url=None, all_queries=False, all_mutations=False, query=None, mutation=None, stuff_to_print=None,
        verbose=False):
    if verbose:
        import logging
        log.logger.setLevel(logging.DEBUG)
    try:
        run(file_, url, all_queries, all_mutations, query, mutation, stuff_to_print)
    except Exception:
        import traceback
        traceback.print_exc()
        sys.exit()

    sys.exit(0)


def run(file_, url, all_queries, all_mutations, query, mutation, stuff_to_print):
    schema = parse_schema(file_, url)

    if stuff_to_print:
        print_available_stuff(schema, stuff_to_print)
        return

    # if explicit queries (-q) or mutations (-m) provided, they take priority
    if query or mutation:
        all_queries = all_mutations = False
    # otherwise print all queries and mutations, unless one of the -Q / -M flags is specified
    elif not all_queries and not all_mutations:
        all_queries = all_mutations = True

    # print queries
    if all_queries or query:
        if all_queries:
            queries_to_print = schema.query.fields
        else:
            queries_to_print = (schema.query.fields[q] for q in query.split(','))

        for field in queries_to_print:
            print(schema.generate_query(field).str(pad=4))

    # print mutations
    if schema.mutation and (all_mutations or mutation):
        if all_mutations:
            mutations_to_print = schema.mutation.fields
        else:
            mutations_to_print = (schema.mutation.fields[m] for m in mutation.split(','))

        for field in mutations_to_print:
            print(schema.generate_mutation(field).str(pad=4))


def parse_schema(file_, url):
    # Parse GraphQL schema
    if file_:
        response = json.loads(Path(file_).read_text())
        return GQLSchema(json=response)
    elif url:
        return GQLSchema(url=url)
    else:
        log.err("Either file or url should be provided.")
        sys.exit(1)


def print_available_stuff(schema, stuff_to_print):
    # List stuff if that's what we're asked to do
    if stuff_to_print in ('queries', 'all'):
        for query in schema.query.fields:
            # TODO: Would be nice to print a full prototype (with args) here
            print("query %s" % query.name)
    if stuff_to_print in ('mutations', 'all') and schema.mutation:
        for mutation in schema.mutation.fields:
            print("mutation %s" % mutation.name)
